from __future__ import print_function
from getpass import getpass
import six


if six.PY2:
    input = raw_input  # noqa ; pylint: disable=undefined-variable,redefined-builtin
else:
    from builtins import input  # noqa


class DialogRendererBase(object):
    """ Base class for all renderer implementations """

    def __init__(self):
        pass

    def render(self, dialog):
        """
        Renders the given dialog

        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        raise NotImplementedError()

    def render_password_component(self, component, dialog):
        """
        Renders the password component

        :param component: The password component to render
        :type component: leapp.dialogs.components.PasswordComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        raise NotImplementedError()

    def render_text_component(self, component, dialog):
        """
        Renders the text component

        :param component: The text component to render
        :type component: leapp.dialogs.components.TextComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        raise NotImplementedError()

    def render_choice_component(self, component, dialog):
        """
        Renders the choices component

        :param component: The choices component to render
        :type component: leapp.dialogs.components.ChoicesComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        raise NotImplementedError()

    def render_multiple_choice_component(self, component, dialog):
        """
        Renders the multiple choices component

        :param component: The multiple choices component to render
        :type component: leapp.dialogs.components.MultipleChoiceComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        raise NotImplementedError()

    def render_bool_component(self, component, dialog):
        """
        Renders the boolean component for displaying Yes/No questions

        :param component: The boolean component to render
        :type component: leapp.dialogs.components.BooleanComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        raise NotImplementedError()

    def render_number_component(self, component, dialog):
        """
        Renders the number component

        :param component: The number component to render
        :type component: leapp.dialogs.components.NumberComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        raise NotImplementedError()


class CommandlineRenderer(DialogRendererBase):
    """
    CommandlineRenderer implements the handling for commandline user interactions.
    """
    def __init__(self):
        self.getpass = getpass

    def render(self, dialog):
        """
        Renders the given dialog

        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        if dialog.title:
            self._render_title(dialog.title)
        if dialog.reason:
            self._render_reason(dialog.reason)
        for component in dialog.components:
            if component.value is None:
                # Only render questions we don't know the answer of
                component.dispatch(self, dialog)

    @staticmethod
    def _render_title(value):
        """
        Renders a title.

        :param value: Title value to print.
        :return:  None
        """
        print('\n{}:\n{}\n'.format(value, (len(value) + 1) * '='))

    @staticmethod
    def _render_reason(value):
        """
        Prints a reason string.

        :param value: String to print
        :return:
        """
        print(value + '\n')

    def _render_label(self, value, min_label_width, underline=None):
        """
        Actual rendering of the given label. (Printing it to the output)

        :param value: Label to print.
        :param min_label_width: Minimum label length.
        :param underline: If set, it will print the given value in a second line as many times as there are characters
                        in the formatted label. That means if one would pass '-' it will underline the given value.
        :return: None
        """
        line = self._format_label(value, min_label_width).strip()
        print(line)
        if underline:
            print(underline * len(line))

    @staticmethod
    def _format_label(value, min_label_width, default=None):
        """
        Implementation of component label formatting.

        :param value: String label to format
        :param min_label_width: Minimum width for the label. used to calculate the padding required.
        :param default: In case a default value needs to be displayed, this should be set and it will be printed after
                        the label text in square brackets.
        :return: The formatted label string.
        """
        if default is None:
            default = ''
        else:
            default = ' [{}]'.format(default)
        return ('{value: >' + str(min_label_width + 2) + 's}').format(value=value + default + ': ')

    def render_password_component(self, component, dialog):
        """
        Renders the password component

        :param component: The password component to render
        :type component: leapp.dialogs.components.PasswordComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        dialog.answer(component, self.getpass(prompt=self._format_label(component.label or 'Password: ',
                                                                        min_label_width=dialog.min_label_width)))

    def render_text_component(self, component, dialog):
        """
        Renders the text component

        :param component: The text component to render
        :type component: leapp.dialogs.components.TextComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        return self._render_text_component(component, dialog)

    def _render_text_component(self, component, dialog, result_hook=None, default_hint=None):
        """
        Performs the actual rendering of the text component.

        :param component: The text component to render
        :type component: leapp.dialogs.components.TextComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :param result_hook: An optional hook to handle special result values (e.g. bool/integral cases)
        :param default_hint: An optional override of the default_hint to show.
        :return: None
        """
        if default_hint is None:
            default_hint = component.default
        while True:
            result = input(
                self._format_label(component.label, default=default_hint, min_label_width=dialog.min_label_width))
            if not result and component.default is not None:
                result = component.default
            elif result and result_hook:
                if result_hook(component, dialog, result):
                    break
                continue
            elif not result:
                continue
            dialog.answer(component, result)
            break

    def render_choice_component(self, component, dialog):
        """
        Renders the choices component

        :param component: The choices component to render
        :type component: leapp.dialogs.components.ChoicesComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        if not component.choices:
            return
        indices = '0123456789abcdefghijklmnopqrstuvwxyz'
        selected = set()
        if component.multi and component.default:
            selected = {component.choices.index(x) for x in component.default}
        while True:
            self._render_label(component.label, min_label_width=dialog.min_label_width, underline='-')
            for idx, choice in enumerate(component.choices):
                checked = ''
                if component.multi:
                    checked = ' [{}] '.format('X' if idx in selected else ' ')
                print('   {answer}. {checked}{label}'.format(answer=indices[idx], label=choice, checked=checked))
            default = ''
            if component.default is not None:
                if component.multi:
                    default = ' [{}]'.format(tuple(component.choices.index(x) for x in component.default))
                else:
                    default = ' [{}]'.format(component.choices.index(component.default))
            if component.multi:
                label = 'Which entry to toggle (Empty response to continue)'
            else:
                label = 'Choice{default}'.format(default=default)
            result = input(self._format_label(label, len(label)))
            if component.multi:
                if not result:
                    dialog.answer(component, tuple(sorted([component.choices[x] for x in selected])))
                    break
                if len(result) == 1 and -1 < indices.index(result) < len(component.choices):
                    idx = indices.index(result)
                    if idx in selected:
                        selected.remove(idx)
                    else:
                        selected.add(idx)
            else:
                if not result and component.default is not None:
                    dialog.answer(component, component.default)
                    break
                if len(result) == 1 and -1 < indices.index(result) < len(component.choices):
                    dialog.answer(component, component.choices[indices.index(result)])
                    break

    def render_multiple_choice_component(self, component, dialog):
        """
        Renders the multiple choices component

        :param component: The multiple choices component to render
        :type component: leapp.dialogs.components.MultipleChoiceComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        self.render_choice_component(component, dialog)

    def render_bool_component(self, component, dialog):
        """
        Renders the boolean component for displaying Yes/No questions

        :param component: The boolean component to render
        :type component: leapp.dialogs.components.BooleanComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        default_hint = '{}/{}'.format(*component.values)
        if component.default is not None:
            if component.default:
                default_hint = component.values[0]
            else:
                default_hint = component.values[1]
        self._render_text_component(component, dialog, result_hook=self._bool_result_hook,
                                    default_hint=default_hint)

    def render_number_component(self, component, dialog):
        """
        Renders the number component

        :param component: The number component to render
        :type component: leapp.dialogs.components.NumberComponent
        :param dialog: The dialog to render
        :type dialog: leapp.dialogs.dialog.Dialog
        :return: None
        """
        self._render_text_component(component, dialog, result_hook=self._int_result_hook)

    @staticmethod
    def _int_result_hook(component, dialog, result):
        """
        Special hook to handle integral value results

        :param component: Component that is being currently handled
        :param dialog: Dialog the component belongs to
        :param result: Value of the result
        :return: True if the value was valid
        """
        try:
            dialog.answer(component, int(result))
            return True
        except ValueError:
            return False

    @staticmethod
    def _bool_result_hook(component, dialog, result):
        """
        Special hook to handle boolean value results

        :param component: Component that is being currently handled
        :param dialog: Dialog the component belongs to
        :param result: Value of the result
        :return: True if the value was valid
        """

        true_value = component.values[0].lower()
        false_value = component.values[1].lower()

        valid_values = (true_value, false_value)
        true_values = (true_value,)

        if false_value[0] != true_value[0]:
            valid_values += (false_value[0], true_value[0])
            true_values += (true_value[0],)

        result = result.lower()
        valid = result in valid_values
        if valid:
            dialog.answer(component, result in true_values)
        return valid



import sys

from leapp.utils.clicmd import command_arg, command
from leapp.snactor.commands.repo import new_repository

_LONG_DESCRIPTION = '''
DEPRECATED: Please use `snactor repo new` instead

Creates a new local repository for writing Actors, Models, Tags,
Topics, and Workflows or adding shared files, tools or libraries.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@command('new-project', help='[DEPRECATED] Creates a new repository', description=_LONG_DESCRIPTION)
@command_arg('name')
def cli(args):
    sys.stderr.write('WARNING: This command has been deprecated. Please use `snactor repo new`\n')
    new_repository(args)



from leapp.utils.clicmd import command

_LONG_DESCRIPTION = '''
Leapp Workflow related commands.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@command('workflow', help='Workflow related commands', description=_LONG_DESCRIPTION)
def workflow(*args):  # noqa; pylint: disable=unused-argument
    pass



import pytest

from leapp.messaging.inprocess import InProcessMessaging, BaseMessaging
from leapp.models.error_severity import ErrorSeverity
from leapp.models import ErrorModel
from leapp.exceptions import CannotConsumeErrorMessages

from test_models import UnitTestModel
from test_tags import TestTag


class UnitTestModelUnused(UnitTestModel):
    name = 'UnitTestModelUnused'


class FakeActor(object):
    name = 'Fake-actor'
    consumes = (UnitTestModel,)
    produces = ()
    description = '''No description for a fake actor.'''
    tags = (TestTag.Common,)
    apis = ()

    def process(self):
        pass


@pytest.mark.parametrize('stored', (True, False))
def test_messaging_messages(repository_dir, stored):
    with repository_dir.as_cwd():
        msg = InProcessMessaging(stored=stored)
        v = UnitTestModel()
        msg.produce(v, FakeActor())
        consumed = tuple(msg.consume(FakeActor(), UnitTestModel))
        assert len(consumed) == 1
        assert len(msg.messages()) == 1
        assert consumed[0] == v


def test_loading(repository_dir):
    with repository_dir.as_cwd():
        msg = InProcessMessaging()
        with pytest.raises(CannotConsumeErrorMessages):
            msg.load((ErrorModel,))
        msg.load((UnitTestModel,))
        v = UnitTestModel()
        consumed = tuple(msg.consume(FakeActor(), UnitTestModel))
        assert len(consumed) == 1
        assert not msg.messages()
        assert consumed[0] == v

        consumed = tuple(msg.consume(FakeActor(), UnitTestModelUnused))
        assert not consumed
        assert not msg.messages()

        consumed = tuple(msg.consume(FakeActor()))
        assert len(consumed) == 1
        assert not msg.messages()
        assert consumed[0] == v

        consumed = tuple(msg.consume(FakeActor(), UnitTestModelUnused))
        assert not consumed
        assert not msg.messages()


def test_report_error(repository_dir):
    with repository_dir.as_cwd():
        msg = InProcessMessaging()
        msg.report_error('Some error', ErrorSeverity.ERROR, FakeActor(), details=None)
        msg.report_error('Some error with details', ErrorSeverity.ERROR, FakeActor(), details={'foo': 'bar'})
        assert len(msg.errors()) == 2


@pytest.mark.parametrize('stored', (True, False))
def test_not_implemented(repository_dir, stored):
    with repository_dir.as_cwd():
        msg = BaseMessaging(stored=stored)
        if stored:
            with pytest.raises(NotImplementedError):
                msg.report_error('Some error', ErrorSeverity.ERROR, FakeActor(), details=None)
            with pytest.raises(NotImplementedError):
                msg.produce(UnitTestModel(), FakeActor())
        with pytest.raises(NotImplementedError):
            msg.load(FakeActor.consumes)
        msg.consume(FakeActor())
        msg.consume(FakeActor(), UnitTestModel)
        assert not msg.errors()
        assert not msg.messages()
        assert msg.stored == stored



import hashlib
import json
import os

from leapp.reporting import (
    _DEPRECATION_FLAGS,
    Groups,
    Remediation,
    Severity,
    create_report_from_error,
    create_report_from_deprecation
)
from leapp.utils.audit import get_messages, get_audit_entry


def _create_reports_from_deprecations(context_id):
    reports = []
    cache = set()
    for entry in get_audit_entry(event='deprecation', context=context_id):
        data = json.loads(entry['data'])

        # Drop duplicates
        _data_dump = json.dumps(data, sort_keys=True).encode('utf-8')
        data_hash = hashlib.sha256(_data_dump).hexdigest()
        if data_hash in cache:
            continue
        cache.add(data_hash)

        # Create the report
        report = create_report_from_deprecation(data)

        sha256 = hashlib.sha256()
        sha256.update(data_hash.encode('utf-8'))
        sha256.update(entry['context'].encode('utf-8'))

        envelope = {
            'timeStamp': entry['stamp'],
            'hostname': os.environ['LEAPP_HOSTNAME'],
            'actor': entry['actor'],
            'id': sha256.hexdigest()
        }
        report.update(envelope)
        reports.append(report)
    return reports


def fetch_upgrade_report_messages(context_id):
    """
    :param context_id: ID to identify the needed messages
    :type context_id: str
    :return: All upgrade messages of type "Report" withing the given context
    """
    report_msgs = get_messages(names=['Report', 'ErrorModel'], context=context_id) or []

    messages = []
    for message in report_msgs:
        data = message['message']['data']

        # We need to be able to uniquely identify each message so we compute
        # a hash of: context UUID, message ID in the database and hash of the
        # data section of the message itself
        sha256 = hashlib.sha256()
        sha256.update(message['message']['hash'].encode('utf-8'))
        sha256.update(message['context'].encode('utf-8'))
        sha256.update(str(message['id']).encode('utf-8'))

        envelope = {
            'timeStamp': message['stamp'],
            'hostname': message['hostname'],
            'actor': message['actor'],
            'id': sha256.hexdigest()
        }
        data_json = json.loads(data)
        report = json.loads(data_json.get('report', "{}"))
        if not report:
            # transform Error message to Report message
            report = create_report_from_error(data_json)
        report.update(envelope)
        messages.append(report)

    messages.extend(_create_reports_from_deprecations(context_id))
    return messages


SEVERITY_LEVELS = {
    Severity.HIGH: 1,
    Severity.MEDIUM: 2,
    Severity.LOW: 3,
    Severity.INFO: 4
}


def is_inhibitor(message):
    """
    A helper function to check if a message is inhibiting upgrade.
    It takes care of possible differences in report schema, like tags/flags -> groups merge in 1.2.0
    """
    # NOTE(ivasilev) As groups and flags are mutual exclusive, let's not bother with report_schema version
    # and just check both fields for inhibitor presence
    return Groups.INHIBITOR in message.get('groups', message.get('flags', []))


def importance(message):
    if is_inhibitor(message):
        return 0
    return SEVERITY_LEVELS.get(message['severity'], 99)


def generate_report_file(messages_to_report, context, path, report_schema='1.1.0'):
    # NOTE(ivasilev) Int conversion should not break as only specific formats of report_schema versions are allowed
    report_schema_tuple = tuple(int(x) for x in report_schema.split('.'))
    if path.endswith(".txt"):
        with open(path, 'w') as f:
            for message in sorted(messages_to_report, key=importance):
                f.write('Risk Factor: {}{}\n'.format(message['severity'],
                                                     ' (inhibitor)' if is_inhibitor(message) else ''))
                f.write('Title: {}\n'.format(message['title']))
                f.write('Summary: {}\n'.format(message['summary']))
                remediation = Remediation.from_dict(message.get('detail', {}))
                if remediation:
                    f.write('Remediation: {}\n'.format(remediation))
                if report_schema_tuple > (1, 0, 0):
                    # report-schema 1.0.0 doesn't have a stable report key
                    f.write('Key: {}\n'.format(message['key']))
                f.write('-' * 40 + '\n')
    elif path.endswith(".json"):
        with open(path, 'w') as f:
            # Here all possible convertions will take place
            if report_schema_tuple < (1, 1, 0):
                # report-schema 1.0.0 doesn't have a stable report key
                # copy list of messages here not to mess the initial structure for possible further invocations
                messages_to_report = list(messages_to_report)
                for m in messages_to_report:
                    m.pop('key')
            if report_schema_tuple < (1, 2, 0):
                # groups were introduced in 1.2.0, before that there was a tags/flags split
                messages_to_report = list(messages_to_report)
                for msg in messages_to_report:
                    groups = msg.pop('groups', [])
                    msg['flags'] = [g for g in groups if g in _DEPRECATION_FLAGS]
                    msg['tags'] = [g for g in groups if g not in _DEPRECATION_FLAGS]
            if report_schema_tuple == (1, 2, 0):
                # DEPRECATED: flags, tags
                # NOTE(pstodulk): This is a temporary solution to ensure that
                # flags are not present in the JSON output as we need to re-introduce
                # flags internally to ensure that any inhibitor checks will not
                # be negatively affected until we drop flags and tags completely.
                messages_to_report = list(messages_to_report)
                for msg in messages_to_report:
                    # NOTE(ivasilev) As flags field is created only if there is an inhibitor
                    # a default value to pop has to be specified not to end up with a KeyError
                    msg.pop('flags', None)
            json.dump({'entries': messages_to_report, 'leapp_run_id': context}, f, indent=2)



"""
This module implements some testing APIs for API depdency checks. Ensuring that all consumes and produces are still in
place and the code keeps working as expected.

DepCheckAPI4 depends on DepCheckAPI3
DepCheckAPI3 depends on DepCheckAPI2
DepCheckAPI2 depends on DepCheckAPI1

DepCheckAPI4 Produces: DepCheck4 Inherits: Produces: DepCheck2, Consumes DepCheck1, DepCheck3
   |
   L> DepCheckAPI3 Consumes: DepCheck3 Inherits: Produces: DepCheck2, Consumes DepCheck1
         |
         L> DepCheckAPI2 Produces: DepCheck2 Inherits: Consumes DepCheck1
               |
               L> DepCheckAPI1 Consumes: DepCheck1

"""

from leapp.libraries.stdlib import api
from leapp.models import DepCheck1, DepCheck2, DepCheck3, DepCheck4
from leapp.workflows.api import WorkflowAPI


class DepCheckAPI1(WorkflowAPI):
    consumes = (DepCheck1,)

    def consume(self):
        """ Consumes all DepCheck1 messages and returns them as a list """
        return list(api.consume(DepCheck1))


class DepCheckAPI2(WorkflowAPI):
    apis = (DepCheckAPI1,)
    produces = (DepCheck2,)

    def produce(self):
        """ Produces a DepCheck2 message """
        api.produce(DepCheck2())


class DepCheckAPI3(WorkflowAPI):
    apis = (DepCheckAPI2,)
    # Be explicit about what you consume even though DepCheckAPI1 consumes it already and is a dependency
    consumes = (DepCheck1, DepCheck3)

    def consume(self):
        """ Consumes all DepCheck3 messages and returns them as a list """
        return list(api.consume(DepCheck1, DepCheck3))


class DepCheckAPI4(WorkflowAPI):
    apis = (DepCheckAPI3,)
    produces = (DepCheck4,)

    def produce(self):
        """ Produces a DepCheck4 message """
        api.produce(DepCheck4())



"""
This module implements a convenience API for actions that are accessible to actors.

Any code that wants use this convenience library has to be called from within the actors context.
This is true for actors, actor private libraries and repository libraries.
"""
import logging

from leapp.actors import Actor


ErrorSeverity = Actor.ErrorSeverity


def current_actor():
    """
    Retrieve the Actor class instance of the current active actor.
    :return: Instance of the currently instantiated actor.
    :rtype: leapp.actors.Actor
    """
    return Actor.current_instance


def report_error(message, severity=ErrorSeverity.ERROR, details=None):
    """
    Reports an execution error

    :param message: A message to print the possible error
    :type message: str
    :param severity: Severity of the error default :py:attr:`leapp.messaging.errors.ErrorSeverity.ERROR`
    :type severity: str with defined values from :py:attr:`leapp.messaging.errors.ErrorSeverity.ERROR`
    :param details: A dictionary where additional context information is passed along with the error
    :type details: dict
    :return: None
    """
    return current_actor().report_error(message=message, severity=severity, details=details)


def show_message(message):
    """
    Display a message in user interterface currently in use (CLI, GUI).

    It uses one of the dialog renderers in :py:mod:`leapp.dialogs.renderer`.

    :param message: Message to show
    :type message: str
    """
    return current_actor().show_message(message=message)


def current_logger():
    """
    Retrieve the logger of the current active actor.
    :return: Logger instance for the current actor.
    :rtype: logging.Logger
    """
    return current_actor().log if current_actor() else logging.getLogger('leapp.fallback')


def produce(*model_instances):
    """
    By calling produce, model instances are stored as messages. Those messages can be then consumed by other actors.

    :param model_instances: Messages to be sent (those model types have to be specified in :py:attr:`produces`
    :type model_instances: Variable number of instances of derived classes from :py:class:`leapp.models.Model`
    """
    return current_actor().produce(*model_instances)


def consume(*models):
    """
    Retrieve messages specified in the actors :py:attr:`consumes` attribute, and filter message types by
    models.

    :param models: Models to use as a filter for the messages to return
    :type models: Variable number of the derived classes from :py:class:`leapp.models.Model`
    """
    return current_actor().consume(*models)


def get_answers(dialog):
    """
    Get the answers for a dialog from answerfile. The dialog needs be predefined in :py:attr:`dialogs` of the actor.

    :param dialog: Dialog instance to show
    :return: dictionary with the requested answers, None if not a defined dialog
    """
    return current_actor().get_answers(dialog)


def actor_files_paths():
    """
    Returns the file paths that are bundled with the actor. (Path to the content of the actor's file directory).
    """
    return current_actor().actor_files_paths


def files_paths():
    """ Returns all actor file paths related to the actor and common actors file paths. """
    return current_actor().files_paths


def common_files_paths():
    """ Returns all common repository file paths. """
    return current_actor().common_files_paths


def actor_tools_paths():
    """
    Returns the tool paths that are bundled with the actor. (Path to the content of the actor's tools directory).
    """
    return current_actor().actor_tools_paths


def tools_paths():
    """ Returns all actor tools paths related to the actor and common actors tools paths. """
    return current_actor().tools_paths


def common_tools_paths():
    """ Returns all common repository tool paths. """
    return current_actor().common_tools_paths


def get_common_folder_path(name):
    """
    Finds the first matching folder path within :py:attr:`files_paths`.

    :param name: Name of the folder
    :type name: str
    :return: Found folder path
    :rtype: str or None
    """
    return current_actor().get_common_folder_path(name)


def get_actor_folder_path(name):
    """
    Finds the first matching folder path within :py:attr:`files_paths`.

    :param name: Name of the folder
    :type name: str
    :return: Found folder path
    :rtype: str or None
    """
    return current_actor().get_actor_folder_path(name)


def get_folder_path(name):
    """
    Finds the first matching folder path within :py:attr:`files_paths`.

    :param name: Name of the folder
    :type name: str
    :return: Found folder path
    :rtype: str or None
    """
    return current_actor().get_folder_path(name)


def get_common_file_path(name):
    """
    Finds the first matching file path within :py:attr:`files_paths`.

    :param name: Name of the file
    :type name: str
    :return: Found file path
    :rtype: str or None
    """
    return current_actor().get_common_file_path(name)


def get_actor_file_path(name):
    """
    Finds the first matching file path within :py:attr:`files_paths`.

    :param name: Name of the file
    :type name: str
    :return: Found file path
    :rtype: str or None
    """
    return current_actor().get_actor_file_path(name)


def get_file_path(name):
    """
    Finds the first matching file path within :py:attr:`files_paths`.

    :param name: Name of the file
    :type name: str
    :return: Found file path
    :rtype: str or None
    """
    return current_actor().get_file_path(name)


def get_tool_path(name):
    """
    Finds the first matching executable file path within :py:attr:`tools_paths`.

    :param name: Name of the file
    :type name: str
    :return: Found file path
    :rtype: str or None
    """
    return current_actor().get_tool_path(name)


def get_common_tool_path(name):
    """
    Finds the first matching executable file path within :py:attr:`common_tools_paths`.

    :param name: Name of the file
    :type name: str
    :return: Found file path
    :rtype: str or None
    """
    return current_actor().get_common_tool_path(name)


def get_actor_tool_path(name):
    """
    Finds the first matching executable file path within :py:attr:`actor_tools_paths`.

    :param name: Name of the file
    :type name: str
    :return: Found file path
    :rtype: str or None
    """
    return current_actor().get_actor_tool_path(name)



from leapp.compat import unicode_type


class Component(object):
    """
    Base class for all components
    """
    key = None
    label = None
    description = None
    default = None
    value = None
    reason = None
    value_type = None

    def __init__(self, key=None, label=None, description=None, default=None, reason=None):
        """

        :param key: Unique key within a dialog scope. Needs to be in the format: `[a-zA-Z_][a-zA-Z0-9_]*`
        :param label: Label for the input to print
        :param description: Description what this value is used for.
        :param default: Default value to
        :param reason: The reason why we need this value.
        """
        self.key = key or type(self).key
        self.label = label.replace('\n', ' ').strip() if label else type(self).label
        self.description = description or type(self).description
        self.default = default if default is not None else type(self).default
        self.value = type(self).value
        self.reason = reason or type(self).reason

    def dispatch(self, renderer, dialog):
        raise NotImplementedError()

    def serialize(self):
        """
        :return: Serialized component information
        """
        return {
            'key': self.key,
            'label': self.label,
            'description': self.description,
            'default': self.default,
            'value': self.value,
            'reason': self.reason,
            'value_type': str(self.value_type) if self.value_type else None
        }


class TextComponent(Component):
    """
    TextComponent is a text input component.
    """
    value_type = str

    def dispatch(self, renderer, dialog):
        renderer.render_text_component(self, dialog=dialog)


class PasswordComponent(TextComponent):
    """
    PasswordComponent is a text input component which will use non echoing input when possible (see getpass).
    """

    label = 'Password'
    value_type = str

    def dispatch(self, renderer, dialog):
        renderer.render_password_component(self, dialog=dialog)


class NumberComponent(Component):
    """
    NumberComponent is used for integer inputs.
    """
    value_type = int
    default = -1

    def dispatch(self, renderer, dialog):
        renderer.render_number_component(self, dialog=dialog)


class BooleanComponent(Component):
    """
    BooleanComponent is used for boolean inputs such as Yes/No questions.
    """
    choices = ('True', 'False')
    values = ('Yes', 'No')
    value_type = bool

    def __init__(self, key=None, label=None, description=None, default=None, reason=None, values=None):
        """
        :param key: Unique key within a dialog scope. Needs to be in the format: `[a-zA-Z_][a-zA-Z0-9_]*`
        :param label: Label for the input to print
        :param description: Description what this value is used for.
        :param default: Default value to
        :param reason: The reason why we need this value.
        :param values: Values to use as True and False, first is always True and the second is always False
                       (e.g. Yes/No)
        """
        super(BooleanComponent, self).__init__(key=key, label=label, description=description, default=default,
                                               reason=reason)
        self.values = values or type(self).values

    def dispatch(self, renderer, dialog):
        renderer.render_bool_component(self, dialog=dialog)


class ChoiceComponent(Component):
    """
    ChoiceComponent is used to give a list of options and allows to select one (like a radio button)
    """
    choices = ()
    multi = False
    value_type = unicode_type

    def __init__(self, choices=None, key=None, label=None, description=None, default=None, reason=None):
        """
        :param key: Unique key within a dialog scope. Needs to be in the format: `[a-zA-Z_][a-zA-Z0-9_]*`
        :param label: Label for the input to print
        :param description: Description what this value is used for.
        :param default: Default value to
        :param reason: The reason why we need this value.
        :param choices: Choices that are available to the user
        """
        super(ChoiceComponent, self).__init__(key=key, label=label, description=description, default=default,
                                              reason=reason)
        self.choices = choices or type(self).choices

    def dispatch(self, renderer, dialog):
        renderer.render_choice_component(self, dialog=dialog)


class MultipleChoiceComponent(ChoiceComponent):
    """
    MultipleChoiceComponent is used to give a list of options and allows to select more than one (like checkboxes)
    """
    choices = ()
    multi = True
    value_type = tuple  # indices

    def dispatch(self, renderer, dialog):
        """
        Calls the appropriate rendering implementation on the renderer and passes itself and the dialog instance to it.

        :param renderer: Renderer instance
        :type renderer: :py:class:`leapp.dialogs.renderer.DialogRendererBase`
        :param dialog:
        :return:
        """
        renderer.render_multiple_choice_component(self, dialog=dialog)



from leapp.actors import Actor
from leapp.tags import InplaceUpgradeWorkflowTag, FirstBootTag, ReRunVerifyTag


class ReRunActor(Actor):
    """
    No documentation has been provided for the re_run_actor actor.
    """

    name = 're_run_actor'
    consumes = ()
    produces = ()
    tags = (InplaceUpgradeWorkflowTag, FirstBootTag, ReRunVerifyTag)

    def process(self):
        print('<<<TEST>>>: {}'.format(self.__class__.__name__))



import os

import pytest

from helpers import TESTING_REPOSITORY_NAME
from leapp.exceptions import CommandError
from leapp.utils.repository import requires_repository, to_snake_case, make_class_name, make_name,\
    find_repository_basedir, get_repository_name, get_repository_metadata


def setup_module():
    os.environ['PYTHONDONTWRITEBYTECODE'] = '1'


@requires_repository
def _test_repository_dir_presence():
    pass


def test_requires_repository(repository_dir):
    with pytest.raises(CommandError):
        _test_repository_dir_presence()
    with repository_dir.as_cwd():
        _test_repository_dir_presence()


_SNAKE_CASE_CONVERSION_FIXTURES = (
    ("UpperNameCase", "upper_name_case"),
    ("snake_case_persists", "snake_case_persists"),
    ("Dash-Separated-Names", "dash_separated_names")
)


@pytest.mark.parametrize("param", _SNAKE_CASE_CONVERSION_FIXTURES)
def test_to_snake_case(param):
    assert to_snake_case(param[0]) == param[1]


_MAKE_CLASS_NAME_CONVERSION_FIXTURES = (
    ("UpperNameCase", "UpperNameCase"),
    ("snake_case_modified", "SnakeCaseModified"),
    ("Dash-Separated-Names", "DashSeparatedNames"),
    ("dash-separated-names-lower", "DashSeparatedNamesLower"),
)


@pytest.mark.parametrize("param", _MAKE_CLASS_NAME_CONVERSION_FIXTURES)
def test_make_class_name(param):
    assert make_class_name(param[0]) == param[1]


@pytest.mark.parametrize("param", _SNAKE_CASE_CONVERSION_FIXTURES)
def test_make_name(param):
    assert make_name(param[0]) == param[1]


def test_find_repository_basedir(repository_dir):
    nested = repository_dir.mkdir('a').mkdir('b').mkdir('c')
    assert repository_dir.samefile(find_repository_basedir(nested.strpath))
    assert repository_dir.samefile(find_repository_basedir(repository_dir.strpath))
    assert find_repository_basedir('.') is None


def test_get_repository_metadata(repository_dir):
    assert get_repository_name(repository_dir.strpath) == TESTING_REPOSITORY_NAME
    assert get_repository_metadata(repository_dir.strpath)['name'] == TESTING_REPOSITORY_NAME
    assert not get_repository_metadata('.')



from leapp.exceptions import TagFilterUsageError
from leapp.tags import Tag


class TagFilter(object):
    def __init__(self, phase, *tags):
        if not phase or not isinstance(phase, type) or not issubclass(phase, Tag):
            raise TagFilterUsageError("TagFilter phase parameter needs to be set to a tag.")
        self.phase = phase
        self.tags = tags

    def get_before(self):
        result = set(actor for actor in self.phase.Before.actors)
        # tag.actors retrives all actors that have the particular tag
        for tag in self.tags:
            result.intersection_update(tag.actors)
        result.update(self.phase.Before.Common.actors)
        return tuple(result)

    def get_after(self):
        result = set(actor for actor in self.phase.After.actors)
        # tag.actors retrives all actors that have the particular tag
        for tag in self.tags:
            result.intersection_update(tag.actors)
        result.update(self.phase.After.Common.actors)
        return tuple(result)

    def get(self):
        result = set(actor for actor in self.phase.actors)
        # tag.actors retrives all actors that have the particular tag
        for tag in self.tags:
            result.intersection_update(tag.actors)
        result.update(self.phase.Common.actors)
        return tuple(result)

    def serialize(self):
        """
        :return: Serialized data for tag filter
        """
        return {
            'phase': self.phase.__name__,
            'tags': [tag.__name__ for tag in self.tags],
        }



from leapp.workflows import Workflow
from leapp.workflows.phases import Phase
from leapp.workflows.flags import Flags
from leapp.workflows.tagfilters import TagFilter
from leapp.workflows.policies import Policies
from leapp.tags import DeprecationWorkflowTag, DeprecationPhaseTag


class DeprecationWorkflow(Workflow):
    name = 'DeprecationWorkflow'
    tag = DeprecationWorkflowTag
    short_name = 'deprecation_workflow'
    description = '''No description has been provided for the DeprecationWorkflow workflow.'''

    class PhaseName(Phase):
        name = 'deprecation_phase'
        filter = TagFilter(DeprecationPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()



import os
import sys

from subprocess import check_call

from leapp.snactor.commands.workflow import workflow
from leapp.utils.clicmd import command_arg, command_opt
from leapp.utils.repository import requires_repository, make_class_name, make_name, find_repository_basedir

_LONG_DESCRIPTION = '''
Creates a new workflow with the given name.

Short name can be overridden using the --short-name option. The class name
can be overridden by using the --class-name option.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@workflow.command('new', help='Creates a new workflow with the given name', description=_LONG_DESCRIPTION)
@command_arg('name')
@command_opt('short-name', short_name='s', help='Used short name for the workflow')
@command_opt('class-name', short_name='c', help='Used class name of the workflow')
@requires_repository
def cli(args):
    class_name = args.class_name
    short_name = args.short_name
    name = args.name
    base_dir = find_repository_basedir('.')
    workflows_dir = os.path.join(base_dir, 'workflows')

    class_name = class_name or make_class_name(name)
    short_name = short_name or make_name(name)
    if not class_name.endswith('Workflow'):
        class_name += 'Workflow'

    check_call(['snactor', 'new-tag', class_name])

    if not os.path.exists(workflows_dir):
        os.mkdir(workflows_dir)

    workflow_path = os.path.join(workflows_dir, make_name(name) + '.py')
    if not os.path.exists(workflow_path):
        with open(workflow_path, 'w') as f:
            f.write("""from leapp.workflows import Workflow
from leapp.workflows.phases import Phase
from leapp.workflows.flags import Flags
from leapp.workflows.tagfilters import TagFilter
from leapp.workflows.policies import Policies
from leapp.tags import {workflow_class}Tag


class {workflow_class}(Workflow):
    name = '{workflow_name}'
    tag =  {workflow_class}Tag
    short_name = '{workflow_short_name}'
    description = '''No description has been provided for the {workflow_name} workflow.'''

    # Template for phase definition - The order in which the phase classes are defined
    # within the Workflow class represents the execution order
    #
    # class PhaseName(Phase):
    #    name = 'phase_name'
    #    filter = TagFilter(PhaseTag)
    #    policies = Policies(Policies.Errors.FailPhase,
    #                        Policies.Retry.Phase)
    #    flags = Flags()
""".format(workflow_name=name, workflow_class=class_name, workflow_short_name=short_name))
    sys.stdout.write("New workflow {} has been created in {}\n".format(class_name,
                                                                       os.path.realpath(workflow_path)))



import pytest

import leapp.models
import leapp.models.fields
from leapp.models.utils import init_from_tuple
from leapp.exceptions import ModelDefinitionError
from test_topics import UnitTestTopic


class UnitTestModel(leapp.models.Model):
    topic = UnitTestTopic
    strings = leapp.models.fields.Nullable(leapp.models.fields.List(leapp.models.fields.String()))
    integer = leapp.models.fields.Nullable(leapp.models.fields.Integer())
    items = leapp.models.fields.List(leapp.models.fields.String(), default=[])


class InheritedUnitTestModel(UnitTestModel):
    pass


class AnotherInheritedUnitTestModel(UnitTestModel):
    pass


class ExtendedInheritedUnitTestModel(InheritedUnitTestModel):
    boolean = leapp.models.fields.Boolean(default=False)


class ExtendedOverriddenInheritedUnitTestModel(ExtendedInheritedUnitTestModel):
    integer = leapp.models.fields.Number()


def test_model_definition_error():
    with pytest.raises(ModelDefinitionError):
        type('FailingModelDefinition', (leapp.models.Model,), {})


def test_get_models():
    models = leapp.models.get_models()
    assert len(models) >= 2
    assert UnitTestModel in models
    assert leapp.models.ErrorModel in models


def test_init_from_tuple():
    strings_value = ['first', 'second', 'third']
    integer_value = 2
    model = init_from_tuple(UnitTestModel, ('strings', 'integer'), (strings_value, integer_value))
    assert model.strings == ['first', 'second', 'third'] and model.integer == integer_value


def test_inheritance():
    strings_value = ['first', 'second', 'third']
    integer_value = 2.4
    boolean_value = True
    a = ExtendedOverriddenInheritedUnitTestModel(strings=strings_value,
                                                 integer=integer_value,
                                                 boolean=boolean_value)
    assert a.boolean == boolean_value
    assert a.strings == strings_value
    assert a.integer == integer_value
    assert len(ExtendedOverriddenInheritedUnitTestModel.fields) == 4

    with pytest.raises(leapp.models.fields.ModelViolationError):
        # Ensure that passing a wrong value type will raise an exception if the field was defined in a base class
        # and overridden by a derived class. In this case  ExtendedOverriddenInheritedUnitTestModel overrides the field
        # integer to be a number, passing a number to the non overridden base class will still keep raising an exception
        ExtendedInheritedUnitTestModel(strings=strings_value, integer=integer_value, boolean=boolean_value)

    assert isinstance(ExtendedOverriddenInheritedUnitTestModel.fields['boolean'], leapp.models.fields.Boolean)
    assert isinstance(ExtendedOverriddenInheritedUnitTestModel.fields['integer'], leapp.models.fields.Number)
    assert isinstance(ExtendedOverriddenInheritedUnitTestModel.fields['strings'], leapp.models.fields.List)

    assert isinstance(ExtendedInheritedUnitTestModel.fields['boolean'], leapp.models.fields.Boolean)
    assert isinstance(ExtendedInheritedUnitTestModel.fields['integer'], leapp.models.fields.Integer)
    assert isinstance(ExtendedInheritedUnitTestModel.fields['strings'], leapp.models.fields.List)

    assert len(InheritedUnitTestModel.fields) == 3
    assert isinstance(InheritedUnitTestModel.fields['integer'], leapp.models.fields.Integer)
    assert isinstance(InheritedUnitTestModel.fields['strings'], leapp.models.fields.List)
    assert InheritedUnitTestModel.fields is not UnitTestModel.fields


def test_inheritance_list_field_not_shared():
    model1 = InheritedUnitTestModel()
    model2 = AnotherInheritedUnitTestModel()
    model3 = InheritedUnitTestModel()
    model4 = AnotherInheritedUnitTestModel()
    model1.items.append('Drogon')
    model2.items.append('Viserion')
    # check for value
    assert model1.items != model2.items
    assert model3.items == model4.items
    # check for identity
    models = [model1, model2, model3, model4]
    for i, model_i in enumerate(models):
        for model_j in models[i + 1:]:
            assert model_i.items is not model_j.items



from contextlib import contextmanager
import datetime
import json
import os
import sqlite3

from leapp.config import get_config
from leapp.compat import string_types
from leapp.utils.schemas import CURRENT_SCHEMA, MIGRATIONS


@contextmanager
def _umask(new_mask):
    cur_mask = os.umask(new_mask)
    yield
    os.umask(cur_mask)


def _initialize_database(db):
    """
    Checks that the latest schema of the database has been applied or initializes the database
    for new databases

    :param db: Connection object to the database
    :return: Pass through of the db object for convenience
    """
    schema_version = db.execute('PRAGMA schema_version').fetchone()[0]
    if not schema_version:
        db.executescript(CURRENT_SCHEMA)
    else:
        user_version = db.execute('PRAGMA user_version').fetchone()[0]
        versions = [m[0] for m in MIGRATIONS]
        try:
            index = versions.index(user_version)
        except ValueError:
            pass
        else:
            for migration in MIGRATIONS[index:]:
                db.executescript(migration[1])

    if os.environ.get('LEAPP_DEVEL_DATABASE_SYNC_OFF'):
        db.execute('PRAGMA journal_mode = WAL')
        # This code speeds up leapp by a factor of almost 18 in some scenarios
        # however comes at the cost of potential database corruption
        # According to the SQLITE documentation this corruption can only happen
        # in case of power loss or an OS crash.
        db.execute('PRAGMA synchronous = 0')

    return db


def create_connection(path):
    """
    Creates a database connection to the path and ensures it's initialized and up to date.

    :param path: Path to the database
    :return: Connection object
    """
    with _umask(0o177):
        return _initialize_database(sqlite3.connect(path))


def get_connection(db):
    """
    Get the database connection or passes it through if it is already set

    :param db: Database connection to be passed through in case it exists already
    :return: database object initialized and migrated to the latest schema version
    """
    if db:
        return db
    cfg = get_config()
    return create_connection(cfg.get('database', 'path'))


class Storable(object):
    """
    Base class for database storables
    """

    def store(self, db=None):
        """
        Stores the data within a transaction
        :param db: Database object (optional)
        :return: None
        """
        with get_connection(db) as connection:
            self.do_store(connection)

    def do_store(self, connection):
        """
        Performs the actual storing of the data

        :param connection: Database connection to use (Can be a transaction cursor)
        :return: None
        """


class Execution(Storable):
    """
    Stores information about the current execution
    """

    @property
    def execution_id(self):
        """
        Returns the id of the entry, which is only set when already stored.
        :return: Integer id or None
        """
        return self._execution_id

    def __init__(self, context=None, kind=None, configuration=None, stamp=None):
        """
        :param context: Execution context identifier
        :type context: str
        :param kind: Execution kind - Can be any string and used for filtering
        :type kind: str
        :param configuration:
        :type configuration: str, dict, list or tuple
        :param stamp: Timestamp string of the execution start in iso format
        :type stamp: str
        """
        super(Execution, self).__init__()
        self.stamp = stamp or datetime.datetime.utcnow().isoformat() + 'Z'
        if not isinstance(configuration, string_types):
            configuration = json.dumps(configuration, sort_keys=True)
        self.configuration = configuration
        self.context = context
        self.kind = kind
        self._execution_id = None

    def do_store(self, connection):
        super(Execution, self).do_store(connection)
        connection.execute('INSERT OR IGNORE INTO execution (context, configuration, stamp, kind) VALUES(?, ?, ?, ?)',
                           (self.context, self.configuration, self.stamp, self.kind))
        cursor = connection.execute('SELECT id FROM execution WHERE context = ?',
                                    (self.context,))
        self._execution_id = cursor.fetchone()[0]


class Host(Storable):
    """
    Host information
    """

    def __init__(self, context=None, hostname=None):
        self.context = context
        self.hostname = hostname
        self._host_id = None

    @property
    def host_id(self):
        """
        Returns the id of the entry, which is only set when already stored.
        :return: Integer id or None
        """
        return self._host_id

    def do_store(self, connection):
        super(Host, self).do_store(connection)
        connection.execute('INSERT OR IGNORE INTO host (context, hostname) VALUES(?, ?)',
                           (self.context, self.hostname))
        cursor = connection.execute('SELECT id FROM host WHERE context = ? AND hostname = ?',
                                    (self.context, self.hostname))
        self._host_id = cursor.fetchone()[0]


class MessageData(Storable):
    """
    Message data
    """

    def __init__(self, data=None, hash_id=None):
        """
        :param data: Message payload
        :type data: str
        :param hash_id: SHA256 hash in hexadecimal representation of data
        :type hash_id: str
        """
        super(MessageData, self).__init__()
        self.data = data
        self.hash_id = hash_id

    def do_store(self, connection):
        super(MessageData, self).do_store(connection)
        connection.execute('INSERT OR IGNORE INTO message_data (hash, data) VALUES(?, ?)', (self.hash_id, self.data))


class DataSource(Host):
    def __init__(self, actor=None, phase=None, context=None, hostname=None):
        """
        :param actor: Name of the actor that triggered the entry
        :type actor: str
        :param phase: In which phase of the workflow execution the data entry was created
        :type phase: str
        :param context: The execution context
        :type context: str
        :param hostname: Hostname of the system that produced the entry
        :type hostname: str
        """
        super(DataSource, self).__init__(context=context, hostname=hostname)
        self.actor = actor
        self.phase = phase
        self._data_source_id = None

    @property
    def data_source_id(self):
        """
        Returns the id of the entry, which is only set when already stored.
        :return: Integer id or None
        """
        return self._data_source_id

    def do_store(self, connection):
        super(DataSource, self).do_store(connection)
        connection.execute('INSERT OR IGNORE INTO data_source (context, host_id, actor, phase) VALUES(?, ?, ?, ?)',
                           (self.context, self.host_id, self.actor, self.phase))
        cursor = connection.execute(
            'SELECT id FROM data_source WHERE context = ? AND host_id = ? AND actor = ? AND phase = ?',
            (self.context, self.host_id, self.actor, self.phase))
        self._data_source_id = cursor.fetchone()[0]


class Message(DataSource):
    def __init__(self, stamp=None, msg_type=None, topic=None, data=None, actor=None, phase=None,
                 hostname=None, context=None):
        """
        :param stamp: Timestamp string of the message creation in iso format
        :type stamp: str
        :param msg_type: Name of the model that represents the message payload
        :type msg_type: str
        :param topic: Topic for this message
        :type topic: str
        :param data: Payload data
        :type data: :py:class:`leapp.utils.audit.MessageData`
        :param actor: Name of the actor that triggered the entry
        :type actor: str
        :param phase: In which phase of the workflow execution the message was created
        :type phase: str
        :param context: The execution context
        :type context: str
        :param hostname: Hostname of the system that produced the message
        :type hostname: str
        """
        super(Message, self).__init__(actor=actor, phase=phase, hostname=hostname, context=context)
        self.stamp = stamp or datetime.datetime.utcnow().isoformat() + 'Z'
        self.msg_type = msg_type
        self.topic = topic
        self.data = data
        self._message_id = None

    @property
    def message_id(self):
        """
        Returns the id of the entry, which is only set when already stored.
        :return: Integer id or None
        """
        return self._message_id

    def do_store(self, connection):
        super(Message, self).do_store(connection)
        self.data.do_store(connection)
        cursor = connection.execute(
            'INSERT INTO message (context, stamp, topic, type, data_source_id, message_data_hash) '
            'VALUES(?, ?, ?, ?, ?, ?)',
            (self.context, self.stamp, self.topic, self.msg_type, self.data_source_id, self.data.hash_id))
        self._message_id = cursor.lastrowid


def create_audit_entry(event, data, message=None):
    """
    Create an audit entry

    :param event: Event type identifier
    :param data: Data related to Type of the event, e.g. a command and its arguments
    :param message: An optional message.
    :return:
    """
    Audit(**{
        'actor': os.environ.get('LEAPP_CURRENT_ACTOR', 'NO-ACTOR-SET'),
        'phase': os.environ.get('LEAPP_CURRENT_PHASE', 'NON-WORKFLOW-EXECUTION'),
        'context': os.environ.get('LEAPP_EXECUTION_ID', 'TESTING-CONTEXT'),
        'hostname': os.environ.get('LEAPP_HOSTNAME', 'TESTING-LEAPP-HOSTNAME'),
        'message': message,
        'event': event,
        'data': data
    }).store()


def get_audit_entry(event, context):
    """
    Retrieve audit entries stored in the database for the given context

    :param context: The execution context
    :type context: str
    :param event: Event type identifier
    :type event: str
    :return: list of dicts with id, time stamp, actor and phase fields
    """
    with get_connection(None) as conn:
        cursor = conn.execute('''
            SELECT
                audit.id          AS id,
                audit.stamp       AS stamp,
                audit.data        AS data,
                audit.context     AS context,
                data_source.actor AS actor,
                host.hostname     AS hostname
              FROM
                audit
              JOIN
                data_source ON data_source.id = audit.data_source_id
              JOIN
                host ON data_source.host_id = host.id
              WHERE
                audit.context = ? AND audit.event = ?
              ORDER BY stamp ASC;
        ''', (context, event))
        cursor.row_factory = dict_factory
        return cursor.fetchall()


class Audit(DataSource):
    def __init__(self, event=None, stamp=None, message=None, data=None, actor=None, phase=None, hostname=None,
                 context=None):
        """
        :param event: Type of this event e.g. new-message or log-message but can be anything
        :type event: str
        :param stamp: Timestamp string of the event creation in iso format
        :type stamp: str
        :param message: A message object, if this audit entry represents a message otherwise None
        :type message: :py:class:`leapp.utils.audit.Message` or None
        :param data: If message is None this has to be the data for the audit entry
        :type data: str or None
        :param actor: Name of the actor that triggered the entry
        :type actor: str
        :param phase: In which phase of the workflow execution the data entry was created
        :type phase: str
        :param context: The execution context
        :type context: str
        :param hostname: Hostname of the system that produced the entry
        :type hostname: str
        """
        super(Audit, self).__init__(actor=actor, phase=phase, hostname=hostname, context=context)
        self.event = event
        self.stamp = stamp or datetime.datetime.utcnow().isoformat() + 'Z'
        self.message = message
        self.data = data
        self._audit_id = None

    @property
    def audit_id(self):
        return self._audit_id

    def do_store(self, connection):
        if self._audit_id:
            return

        super(Audit, self).do_store(connection)

        if self.message and not self.message.message_id:
            self.message.do_store(connection)

        if self.data and not isinstance(self.data, string_types):
            self.data = json.dumps(self.data)

        cursor = connection.execute(
            'INSERT INTO audit (event, stamp, context, data_source_id, message_id, data)'
            ' VALUES(?, ?, ?, ?, ?, ?)',
            (self.event, self.stamp, self.context, self.data_source_id,
             self.message.message_id if self.message else None, self.data))

        self._audit_id = cursor.lastrowid


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


_MESSAGE_QUERY_TEMPLATE = '''
        SELECT
             id, context, stamp, topic, type, actor, phase, message_hash, message_data, hostname
        FROM
             messages_data
        WHERE context = ? AND type IN (%s)'''


def get_messages(names, context, connection=None):
    """
    Queries all messages from the database for the given context and the list of model names
    :param names: List of names that should be messages returned for
    :type names: list or tuple of str
    :param context: Execution id the message should be queried from.
    :param connection: Database connection to use instead of the default one.
    :return: Iterable with messages
    :rtype: iterable
    """
    if not names:
        return ()

    with get_connection(db=connection) as conn:
        cursor = conn.execute(_MESSAGE_QUERY_TEMPLATE % ', '.join('?' * len(names)), (context,) + tuple(names))
        cursor.row_factory = dict_factory
        result = cursor.fetchall()

        # Transform to expected format
        for row in result:
            row['message'] = {'data': row.pop('message_data'), 'hash': row.pop('message_hash')}
        return result


_AUDIT_CHECKPOINT_EVENT = 'checkpoint'


def checkpoint(actor, phase, context, hostname):
    """
    Creates a checkpoint audit entry

    :param actor: Name of the actor that triggered the entry
    :type actor: str
    :param phase: In which phase of the workflow execution the data entry was created
    :type phase: str
    :param context: The execution context
    :type context: str
    :param hostname: Hostname of the system that produced the entry
    :type hostname: str
    :return: None
    """

    audit = Audit(event=_AUDIT_CHECKPOINT_EVENT, actor=actor, phase=phase, hostname=hostname, context=context)
    audit.store()


def get_errors(context):
    """
    Queries all error messages from the database for the given context

    :param context: The execution context
    :type context: str
    :return: List of error messages
    """
    return get_messages(('ErrorModel',), context=context)


def get_checkpoints(context):
    """
    Retrieve all checkpoints stored in the database for the given context

    :param context: The execution context
    :type context: str
    :return: list of dicts with id, timestamp, actor and phase fields
    """
    with get_connection(None) as conn:
        cursor = conn.execute('''
            SELECT
                audit.id          AS id,
                audit.stamp       AS stamp,
                data_source.actor AS actor,
                data_source.phase AS phase
              FROM
                audit
              JOIN
                data_source ON data_source.id = audit.data_source_id
              WHERE
                audit.context = ? AND audit.event = ?
              ORDER BY audit.id ASC;
        ''', (context, _AUDIT_CHECKPOINT_EVENT))
        cursor.row_factory = dict_factory
        return cursor.fetchall()



import os

from leapp.actors import Actor
from leapp.tags import ThirdPhaseTag, UnitTestWorkflowTag


class AfterThirdActor(Actor):
    name = 'after_third_actor'
    description = 'No description has been provided for the after_third_actor actor.'
    consumes = ()
    produces = ()
    tags = (ThirdPhaseTag.After, UnitTestWorkflowTag)

    def process(self):
        from leapp.libraries.common.test_helper import log_execution
        log_execution(self)
        if os.environ.get('AfterThirdActor-ReportError') == '1':
            self.report_error("Unit test requested error")



import sys

from leapp.exceptions import InvalidTopicDefinitionError
from leapp.utils.meta import get_flattened_subclasses, with_metaclass


class TopicMeta(type):
    """
    Meta class for the registration of topics
    """

    def __new__(mcs, name, bases, attrs):
        klass = super(TopicMeta, mcs).__new__(mcs, name, bases, attrs)
        setattr(sys.modules[mcs.__module__], name, klass)
        setattr(klass, 'messages', ())
        return klass


class Topic(with_metaclass(TopicMeta)):
    """ Base class for all :ref:`topics <terminology:topic>`"""

    name = None
    """ Name of the topic in snake case """

    messages = ()
    """
    Tuple of :py:class:`leapp.models.Model` derived classes that are using this topic are automatically added to this
    variable.
    """

    @classmethod
    def serialize(cls):
        return {
            'class_name': cls.__name__,
            'name': cls.name,
            'messages': [m.__name__ for m in cls.messages]
        }


class DialogTopic(Topic):
    """
    A special topic for dialogs shown to user during workflow execution.
    """
    name = 'dialog_topic'


class ErrorTopic(Topic):
    """
    A special topic for errors during the execution.
    """
    name = 'errors'


class ReportTopic(Topic):
    """
    A special topic for reporting purposes.
    """
    name = 'report_topic'


def get_topics():
    """
    :return: All registered :py:class:`leapp.topics.Topic` derived classes
    """
    topics = get_flattened_subclasses(Topic)
    for topic in (topic for topic in topics):
        topic_name = getattr(topic, 'name', None)
        if not topic_name:
            raise InvalidTopicDefinitionError('Topic {} does not contain a name attribute'.format(topic))
    return topics
